
/********************************************
bi_funct.c
copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/* $Log: bi_funct.c,v $
 * Revision 1.9  1996/01/14  17:16:11  mike
 * mawk_flush_all_output() before system()
 *
 * Revision 1.8  1995/08/27  18:13:03  mike
 * fix random number generator to work with longs larger than 32bits
 *
 * Revision 1.7  1995/06/09  22:53:30  mike
 * change a memcmp() to strncmp() to make purify happy
 *
 * Revision 1.6  1994/12/13  00:26:32  mike
 * rt_nr and rt_fnr for run-time error messages
 *
 * Revision 1.5  1994/12/11  22:14:11  mike
 * remove THINK_C #defines.  Not a political statement, just no indication
 * that anyone ever used it.
 *
 * Revision 1.4  1994/12/10  21:44:12  mike
 * fflush builtin
 *
 * Revision 1.3  1993/07/14  11:46:36  mike
 * code cleanup
 *
 * Revision 1.2	 1993/07/14  01:22:27  mike
 * rm SIZE_T
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:08  mike
 * move source to cvs
 *
 * Revision 5.5	 1993/02/13  21:57:18  mike
 * merge patch3
 *
 * Revision 5.4	 1993/01/01  21:30:48  mike
 * split mawk_new_STRING() into mawk_new_STRING and mawk_new_STRING0
 *
 * Revision 5.3.1.2  1993/01/27	 01:04:06  mike
 * minor tuning to str_str()
 *
 * Revision 5.3.1.1  1993/01/15	 03:33:35  mike
 * patch3: safer double to int conversion
 *
 * Revision 5.3	 1992/12/17  02:48:01  mike
 * 1.1.2d changes for DOS
 *
 * Revision 5.2	 1992/07/08  15:43:41  brennan
 * patch2: length returns.  I am a wimp
 *
 * Revision 5.1	 1991/12/05  07:55:35  brennan
 * 1.1 pre-release
 *
*/


#include "mawk.h"
#include "bi_funct.h"
#include "bi_vars.h"
#include "memory.h"
#include "init.h"
#include "files.h"
#include "fin.h"
#include "field.h"
#include "regexp.h"
#include "repl.h"
#include <math.h>


/* statics */
static STRING *PROTO(gsub, (mawk_state_t *,PTR, CELL *, char *, int)) ;
static void PROTO(fplib_err, (mawk_state_t *, char *, double, char *)) ;


/* global for the disassembler */
BI_REC const mawk_bi_funct[] =
{				/* info to load builtins */

   "length",  mawk_bi_length, 0, 1,	/* special must come first */
   "index",   mawk_bi_index, 2, 2,
   "substr",  mawk_bi_substr, 2, 3,
   "sprintf", mawk_bi_sprintf, 1, 255,
   "sin",     mawk_bi_sin, 1, 1,
   "cos",     mawk_bi_cos, 1, 1,
   "atan2",   mawk_bi_atan2, 2, 2,
   "exp",     mawk_bi_exp, 1, 1,
   "log",     mawk_bi_log, 1, 1,
   "int",     mawk_bi_int, 1, 1,
   "sqrt",    mawk_bi_sqrt, 1, 1,
   "rand",    mawk_bi_rand, 0, 0,
   "srand",   mawk_bi_srand, 0, 1,
   "close",   mawk_bi_close, 1, 1,
   "system",  mawk_bi_system, 1, 1,
   "toupper", mawk_bi_toupper, 1, 1,
   "tolower", mawk_bi_tolower, 1, 1,
   "fflush",  mawk_bi_fflush, 0, 1,
   "call",    mawk_bi_call, 1, 255,
   "acall",   mawk_bi_acall, 2, 2,
   "valueof", mawk_bi_valueof, 1, 2,

   (char *) 0, (PF_CP) 0, 0, 0} ;


/* load built-in functions in symbol table */
void
mawk_bi_funct_init(MAWK)
   mawk_state_t *MAWK;
{
   register BI_REC *p ;
   register SYMTAB *stp ;

   /* length is special (posix mawk_bozo) */
   stp = mawk_insert(MAWK, mawk_bi_funct->name) ;
   stp->type = ST_LENGTH ;
   stp->stval.bip = mawk_bi_funct ;

   for (p = mawk_bi_funct + 1; p->name; p++)
   {
      stp = mawk_insert(MAWK, p->name) ;
      stp->type = ST_BUILTIN ;
      stp->stval.bip = p ;
   }

   /* seed rand() off the clock */
   {
      CELL c ;

      c.type = 0 ;  mawk_bi_srand(MAWK, &c) ;
   }

}

/**************************************************
 string builtins (except split (in split.c) and [g]sub (at end))
 **************************************************/

CELL *
mawk_bi_length(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   unsigned len ;

   if (sp->type == 0)  mawk_cellcpy(MAWK, sp, MAWK->field) ;
   else	 sp-- ;

   if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
   len = string(sp)->len ;

   free_STRING(string(sp)) ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) len ;

   return sp ;
}

char *
mawk_str_str(target, key, key_len)
   register char *target ;
   char *key ;
   unsigned key_len ;
{
   register int k = key[0] ;

   switch (key_len)
   {
      case 0:
	 return (char *) 0 ;
      case 1:
	 return strchr(target, k) ;
      case 2:
	 {
	    int k1 = key[1] ;
	    while (target = strchr(target, k))
	       if (target[1] == k1)  return target ;
	       else  target++ ;
	    /*failed*/
	    return (char *) 0 ;
	 }
   }

   key_len-- ;
   while (target = strchr(target, k))
   {
      if (strncmp(target + 1, key + 1, key_len) == 0)  return target ;
      else  target++ ;
   }
   /*failed*/
   return (char *) 0 ;
}



CELL *
mawk_bi_index(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   register int idx ;
   unsigned len ;
   char *p ;

   sp-- ;
   if (TEST2(sp) != TWO_STRINGS)  mawk_cast2_to_s(MAWK, sp) ;

   if (len = string(sp + 1)->len)
      idx = (p = mawk_str_str(string(sp)->str, string(sp + 1)->str, len))
	 ? p - string(sp)->str + 1 : 0 ;

   else				/* index of the empty string */
      idx = 1 ;

   free_STRING(string(sp)) ;
   free_STRING(string(sp + 1)) ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) idx ;
   return sp ;
}

/*  substr(s, i, n)
    if l = length(s)  then get the characters
    from  max(1,i) to min(l,n-i-1) inclusive */

CELL *
mawk_bi_substr(MAWK, sp)
   mawk_state_t *MAWK;
   CELL *sp ;
{
   int n_args, len ;
   register int i, n ;
   STRING *sval ;		 /* substr(sval->str, i, n) */

   n_args = sp->type ;
   sp -= n_args ;
   if (sp->type != C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
   /* don't use < C_STRING shortcut */
   sval = string(sp) ;

   if ((len = sval->len) == 0)	/* substr on null string */
   {
      if (n_args == 3)	cell_destroy(sp + 2) ;
      cell_destroy(sp + 1) ;
      return sp ;
   }

   if (n_args == 2)
   {
      n = MAX__INT ;
      if (sp[1].type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp + 1) ;
   }
   else
   {
      if (TEST2(sp + 1) != TWO_DOUBLES)	 mawk_cast2_to_d(MAWK, sp + 1) ;
      n = d_to_i(sp[2].dval) ;
   }
   i = d_to_i(sp[1].dval) - 1 ;	 /* i now indexes into string */

   if ( i < 0 ) { n += i ; i = 0 ; }
   if (n > len - i)  n = len - i ;

   if (n <= 0)			/* the null string */
   {
      sp->ptr = (PTR) & (MAWK->null_str) ;
      (MAWK->null_str.ref_cnt)++ ;
   }
   else	 /* got something */
   {
      sp->ptr = (PTR) mawk_new_STRING0(MAWK, n) ;
      memcpy(string(sp)->str, sval->str + i, n) ;
   }

   free_STRING(sval) ;
   return sp ;
}

/*
  match(s,r)
  sp[0] holds r, sp[-1] holds s
*/

CELL *
mawk_bi_match(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   char *p ;
   unsigned length ;

   if (sp->type != C_RE)  mawk_cast_to_RE(MAWK, sp) ;
   if ((--sp)->type < C_STRING)	 mawk_cast1_to_s(MAWK, sp) ;

   cell_destroy(RSTART) ;
   cell_destroy(RLENGTH) ;
   RSTART->type = C_DOUBLE ;
   RLENGTH->type = C_DOUBLE ;

   p = mawk_REmatch(string(sp)->str, (sp + 1)->ptr, &length) ;

   if (p)
   {
      sp->dval = (double) (p - string(sp)->str + 1) ;
      RLENGTH->dval = (double) length ;
   }
   else
   {
      sp->dval = 0.0 ;
      RLENGTH->dval = -1.0 ;	 /* posix */
   }

   free_STRING(string(sp)) ;
   sp->type = C_DOUBLE ;

   RSTART->dval = sp->dval ;

   return sp ;
}

CELL *
mawk_bi_toupper(MAWK, sp)
   mawk_state_t *MAWK;
   CELL *sp ;
{
   STRING *old ;
   register char *p, *q ;

   if (sp->type != C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
   old = string(sp) ;
   sp->ptr = (PTR) mawk_new_STRING0(MAWK, old->len) ;

   q = string(sp)->str ; p = old->str ;
   while (*p)
   {
      *q = *p++ ;
      if (*q >= 'a' && *q <= 'z')  *q += 'A' - 'a' ;
      q++ ;
   }
   free_STRING(old) ;
   return sp ;
}

CELL *
mawk_bi_tolower(MAWK, sp)
   mawk_state_t *MAWK;
   CELL *sp ;
{
   STRING *old ;
   register char *p, *q ;

   if (sp->type != C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
   old = string(sp) ;
   sp->ptr = (PTR) mawk_new_STRING0(MAWK, old->len) ;

   q = string(sp)->str ; p = old->str ;
   while (*p)
   {
      *q = *p++ ;
      if (*q >= 'A' && *q <= 'Z')  *q += 'a' - 'A' ;
      q++ ;
   }
   free_STRING(old) ;
   return sp ;
}


/************************************************
  arithemetic builtins
 ************************************************/

static void
fplib_err(MAWK, fname, val, error)
   mawk_state_t *MAWK;
   char *fname ;
   double val;
   char *error ;
{
   mawk_rt_error(MAWK, "%s(%g) : %s", fname, val, error) ;
}


CELL *
mawk_bi_sin(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if ! STDC_MATHERR
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = sin(sp->dval) ;
   return sp ;
#else
   double x;

   errno = 0 ;
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   x = sp->dval ;
   sp->dval = sin(sp->dval) ;
   if (errno)  fplib_err(MAWK, "sin", x, "loss of precision") ;
   return sp ;
#endif
}

CELL *
mawk_bi_cos(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if ! STDC_MATHERR
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = cos(sp->dval) ;
   return sp ;
#else
   double x;

   errno = 0 ;
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   x = sp->dval ;
   sp->dval = cos(sp->dval) ;
   if (errno)  fplib_err(MAWK, "cos", x, "loss of precision") ;
   return sp ;
#endif
}

CELL *
mawk_bi_atan2(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if  !	STDC_MATHERR
   sp-- ;
   if (TEST2(sp) != TWO_DOUBLES)  mawk_cast2_to_d(MAWK, sp) ;
   sp->dval = atan2(sp->dval, (sp + 1)->dval) ;
   return sp ;
#else

   errno = 0 ;
   sp-- ;
   if (TEST2(sp) != TWO_DOUBLES)  mawk_cast2_to_d(MAWK, sp) ;
   sp->dval = atan2(sp->dval, (sp + 1)->dval) ;
   if (errno)  mawk_rt_error(MAWK, "atan2(0,0) : domain error") ;
   return sp ;
#endif
}

CELL *
mawk_bi_log(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if ! STDC_MATHERR
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = log(sp->dval) ;
   return sp ;
#else
   double x;

   errno = 0 ;
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   x = sp->dval ;
   sp->dval = log(sp->dval) ;
   if (errno)  fplib_err(MAWK, "log", x, "domain error") ;
   return sp ;
#endif
}

CELL *
mawk_bi_exp(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if  ! STDC_MATHERR
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = exp(sp->dval) ;
   return sp ;
#else
   double x;

   errno = 0 ;
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   x = sp->dval ;
   sp->dval = exp(sp->dval) ;
   if (errno && sp->dval)  fplib_err(MAWK, "exp", x, "mawk_overflow") ;
   /* on underflow sp->dval==0, ignore */
   return sp ;
#endif
}

CELL *
mawk_bi_int(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = sp->dval >= 0.0 ? floor(sp->dval) : ceil(sp->dval) ;
   return sp ;
}

CELL *
mawk_bi_sqrt(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
#if  ! STDC_MATHERR
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   sp->dval = sqrt(sp->dval) ;
   return sp ;
#else
   double x;

   errno = 0 ;
   if (sp->type != C_DOUBLE)  mawk_cast1_to_d(MAWK, sp) ;
   x = sp->dval ;
   sp->dval = sqrt(sp->dval) ;
   if (errno)  fplib_err(MAWK, "sqrt", x, "domain error") ;
   return sp ;
#endif
}

#ifndef NO_TIME_H
#include <time.h>
#else
#include <sys/types.h>
#endif


/* For portability, we'll use our own random number generator , taken
   from:  Park, SK and Miller KW, "Random Number Generators:
   Good Ones are Hard to Find", CACM, 31, 1192-1201, 1988.
*/


#define		M	0x7fffffff	/* 2^31-1 */
#define		MX	0xffffffff
#define		A	16807
#define	  	Q	127773		/* M/A */
#define	  	R	2836		/* M%A */

#if M == MAX__LONG
#define crank(s)   s = A * (s % Q) - R * (s / Q) ;\
		   if ( s <= 0 ) s += M
#else
/* 64 bit longs */
#define crank(s)	{ unsigned long t = s ;\
			  t = (A * (t % Q) - R * (t / Q)) & MX ;\
			  if ( t >= M ) t = (t+M)&M ;\
			  s = t ;\
			}
#endif


CELL *
mawk_bi_srand(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   CELL c ;

   if (sp->type == 0)		/* seed off clock */
   {
      mawk_cellcpy(MAWK, sp, &MAWK->cseed) ;
      cell_destroy(&MAWK->cseed) ;
      MAWK->cseed.type = C_DOUBLE ;
      MAWK->cseed.dval = (double) time((time_t *) 0) ;
   }
   else	 /* user seed */
   {
      sp-- ;
      /* swap cseed and *sp ; don't need to adjust ref_cnts */
      c = *sp ; *sp = MAWK->cseed ; MAWK->cseed = c ;
   }

   /* The old seed is now in *sp ; move the value in cseed to
     seed in range [1,M) */

   mawk_cellcpy(MAWK, &c, &MAWK->cseed) ;
   if (c.type == C_NOINIT)  mawk_cast1_to_d(MAWK, &c) ;

   MAWK->seed = c.type == C_DOUBLE ? (d_to_i(c.dval) & M) % M + 1 :
      mawk_hash(string(&c)->str) % M + 1 ;
   if( MAWK->seed == M ) MAWK->seed = M-1 ;

   cell_destroy(&c) ;

   /* crank it once so close seeds don't give a close
       first result  */
   crank(MAWK->seed) ;

   return sp ;
}

CELL *
mawk_bi_rand(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   crank(MAWK->seed) ;
   sp++ ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) MAWK->seed / (double) M ;
   return sp ;
}
#undef	 A
#undef	 M
#undef   MX
#undef	 Q
#undef	 R
#undef   crank

/*************************************************
 miscellaneous builtins
 close, system and getline
 fflush
 *************************************************/

CELL *
mawk_bi_close(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   int x ;

   if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
   x = mawk_file_close(MAWK, (STRING *) sp->ptr) ;
   free_STRING(string(sp)) ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) x ;
   return sp ;
}


CELL *
mawk_bi_fflush(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   int ret = 0 ;

   if ( sp->type == 0 )  fflush(stdout) ;
   else
   {
      sp-- ;
      if ( sp->type < C_STRING ) mawk_cast1_to_s(MAWK, sp) ;
      ret = mawk_file_flush(MAWK, string(sp)) ;
      free_STRING(string(sp)) ;
   }

   sp->type = C_DOUBLE ;
   sp->dval = (double) ret ;
   return sp ;
}



#if   HAVE_REAL_PIPES

CELL *
mawk_bi_system(MAWK, sp)
   mawk_state_t *MAWK;
   CELL *sp ;
{
   int pid ;
   unsigned ret_val ;

   if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp) ;

   mawk_flush_all_output(MAWK) ;
   switch (pid = fork())
   {
      case -1:			/* fork failed */

	 mawk_errmsg(MAWK, errno, "could not create a new process") ;
	 ret_val = 127 ;
	 break ;

      case 0:			/* the child */
	 execl(MAWK->shell, MAWK->shell, "-c", string(sp)->str, (char *) 0) ;
	 /* if get here, execl() failed */
	 mawk_errmsg(MAWK, errno, "mawk_execute of %s failed", MAWK->shell) ;
	 fflush(stderr) ;
	 _exit(127) ;

      default:			/* wait for the child */
	 ret_val = mawk_wait_for(MAWK, pid) ;
	 break ;
   }

   cell_destroy(sp) ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) ret_val ;
   return sp ;
}

#endif /* HAVE_REAL_PIPES */



#if   MSDOS


CELL *
mawk_bi_system(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   int retval ;

   if (sp->type < C_STRING)  mawk_cast1_to_s(sp) ;
   retval = DOSexec(string(sp)->str) ;
   free_STRING(string(sp)) ;
   sp->type = C_DOUBLE ;
   sp->dval = (double) retval ;
   return sp ;
}

#endif


/*  getline()  */

/*  if type == 0 :  stack is 0 , target address

    if type == F_IN : stack is F_IN, expr(filename), target address
    if type == PIPE_IN : stack is PIPE_IN, target address, expr(pipename)
*/

CELL *
mawk_bi_getline(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   CELL tc, *cp ;
   char *p ;
   unsigned len ;
   FIN *fin_p ;


   switch (sp->type)
   {
      case 0:
	 sp-- ;
	 if (!MAWK->main_fin)	 mawk_open_main(MAWK) ;

   p = mawk_FINgets(MAWK, MAWK->main_fin, &len);
   if (p == (void *)mawk_FIN_nomore) goto eof;

	 if (!p)  goto eof ;

	 cp = (CELL *) sp->ptr ;
	 if (TEST2(NR) != TWO_DOUBLES)	mawk_cast2_to_d(MAWK, NR) ;
	 NR->dval += 1.0 ;  MAWK->rt_nr++ ;
	 FNR->dval += 1.0 ; MAWK->rt_fnr++ ;
	 break ;

      case F_IN:
	 sp-- ;
	 if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
	 fin_p = (FIN *) mawk_file_find(MAWK, sp->ptr, F_IN) ;
	 free_STRING(string(sp)) ;
	 sp-- ;

	 if (!fin_p)  goto open_failure ;

   p = mawk_FINgets(MAWK, fin_p, &len);
   if (p == (void *)mawk_FIN_nomore) goto eof;

	 if (!p)
	 {
	    mawk_FINsemi_close(MAWK, fin_p) ;
	    goto eof ;
	 }
	 cp = (CELL *) sp->ptr ;
	 break ;

      case PIPE_IN:
	 sp -= 2 ;
	 if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp) ;
	 fin_p = (FIN *) mawk_file_find(MAWK, sp->ptr, PIPE_IN) ;
	 free_STRING(string(sp)) ;

	 if (!fin_p)  goto open_failure ;

   p = mawk_FINgets(MAWK, fin_p, &len);
   if (p == (void *)mawk_FIN_nomore) goto eof;

	 if (!p)
	 {
	    mawk_FINsemi_close(MAWK, fin_p) ;
#if  HAVE_REAL_PIPES
	    /* reclaim process slot */
	    mawk_wait_for(MAWK, 0) ;
#endif
	    goto eof ;
	 }
	 cp = (CELL *) (sp + 1)->ptr ;
	 break ;

      default:
	 mawk_bozo(MAWK, "type in mawk_bi_getline") ;

   }

   /* we've read a line , store it */

   if (len == 0)
   {
      tc.type = C_STRING ;
      tc.ptr = (PTR) & (MAWK->null_str) ;
      MAWK->null_str.ref_cnt++ ;
   }
   else
   {
      tc.type = C_MBSTRN ;
      tc.ptr = (PTR) mawk_new_STRING0(MAWK, len) ;
      memcpy(string(&tc)->str, p, len) ;
   }

   mawk_slow_cell_assign(MAWK, cp, &tc) ;

   cell_destroy(&tc) ;

   sp->dval = 1.0  ;  goto done ;
 open_failure:
   sp->dval = -1.0  ; goto done ;
 eof:
   sp->dval = 0.0 ;		 /* fall thru to done  */

 done:sp->type = C_DOUBLE;
   return sp ;
}

/**********************************************
 sub() and gsub()
 **********************************************/

/* entry:  sp[0] = address of CELL to sub on
	   sp[-1] = substitution CELL
	   sp[-2] = regular expression to match
*/

CELL *
mawk_bi_sub(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   CELL *cp ;			 /* pointer to the replacement target */
   CELL tc ;			 /* build the new string here */
   CELL sc ;			 /* copy of the target CELL */
   char *front, *middle, *back ; /* pieces */
   unsigned front_len, middle_len, back_len ;

   sp -= 2 ;
   if (sp->type != C_RE)  mawk_cast_to_RE(MAWK, sp) ;
   if (sp[1].type != C_REPL && sp[1].type != C_REPLV)
      mawk_cast_to_REPL(MAWK, sp + 1) ;
   cp = (CELL *) (sp + 2)->ptr ;
   /* make a copy of the target, because we won't change anything
     including type unless the match works */
   mawk_cellcpy(MAWK, &sc, cp) ;
   if (sc.type < C_STRING)  mawk_cast1_to_s(MAWK, &sc) ;
   front = string(&sc)->str ;

   if (middle = mawk_REmatch(front, sp->ptr, &middle_len))
   {
      front_len = middle - front ;
      back = middle + middle_len ;
      back_len = string(&sc)->len - front_len - middle_len ;

      if ((sp + 1)->type == C_REPLV)
      {
	 STRING *sval = mawk_new_STRING0(MAWK, middle_len) ;

	 memcpy(sval->str, middle, middle_len) ;
	 mawk_replv_to_repl(MAWK, sp + 1, sval) ;
	 free_STRING(sval) ;
      }

      tc.type = C_STRING ;
      tc.ptr = (PTR) mawk_new_STRING0(
			MAWK, front_len + string(sp + 1)->len + back_len) ;

      {
	 char *p = string(&tc)->str ;

	 if (front_len)
	 {
	    memcpy(p, front, front_len) ;
	    p += front_len ;
	 }
	 if (string(sp + 1)->len)
	 {
	    memcpy(p, string(sp + 1)->str, string(sp + 1)->len) ;
	    p += string(sp + 1)->len ;
	 }
	 if (back_len)	memcpy(p, back, back_len) ;
      }

      mawk_slow_cell_assign(MAWK, cp, &tc) ;

      free_STRING(string(&tc)) ;
   }

   free_STRING(string(&sc)) ;
   mawk_repl_destroy(MAWK, sp + 1) ;
   sp->type = C_DOUBLE ;
   sp->dval = middle != (char *) 0 ? 1.0 : 0.0 ;
   return sp ;
}

/* recursive global subsitution
   dealing with empty matches makes this mildly painful
*/

static STRING *
gsub(MAWK, re, repl, target, flag)
   mawk_state_t *MAWK;
   PTR re ;
   CELL *repl ;			 /* always of type REPL or REPLV,
       destroyed by caller */
   char *target ;

   /* if on, match of empty string at front is OK */
   int flag ;
{
   char *front, *middle ;
   STRING *back ;
   unsigned front_len, middle_len ;
   STRING *ret_val ;
   CELL xrepl ;			 /* a copy of repl so we can change repl */

   if (!(middle = mawk_REmatch(target, re, &middle_len)))
      return mawk_new_STRING(MAWK, target) ;/* no match */

   mawk_cellcpy(MAWK, &xrepl, repl) ;

   if (!flag && middle_len == 0 && middle == target)
   {				/* match at front that's not allowed */

      if (*target == 0)		/* target is empty string */
      {
	 mawk_repl_destroy(MAWK, &xrepl) ;
	 MAWK->null_str.ref_cnt++ ;
	 return &(MAWK->null_str) ;
      }
      else
      {
	 char xbuff[2] ;

	 front_len = 0 ;
	 /* make new repl with target[0] */
	 mawk_repl_destroy(MAWK, repl) ;
	 xbuff[0] = *target++ ;	 xbuff[1] = 0 ;
	 repl->type = C_REPL ;
	 repl->ptr = (PTR) mawk_new_STRING(MAWK, xbuff) ;
	 back = gsub(MAWK, re, &xrepl, target, 1) ;
      }
   }
   else	 /* a match that counts */
   {
      MAWK->repl_cnt++ ;

      front = target ;
      front_len = middle - target ;

      if (*middle == 0)		/* matched back of target */
      {
	 back = &(MAWK->null_str) ;
	 MAWK->null_str.ref_cnt++ ;
      }
      else  back = gsub(MAWK, re, &xrepl, middle + middle_len, 0) ;

      /* patch the &'s if needed */
      if (repl->type == C_REPLV)
      {
	 STRING *sval = mawk_new_STRING0(MAWK, middle_len) ;

	 memcpy(sval->str, middle, middle_len) ;
	 mawk_replv_to_repl(MAWK, repl, sval) ;
	 free_STRING(sval) ;
      }
   }

   /* put the three pieces together */
   ret_val = mawk_new_STRING0(MAWK, front_len + string(repl)->len + back->len) ;
   {
      char *p = ret_val->str ;

      if (front_len)
      {
	 memcpy(p, front, front_len) ;
	 p += front_len ;
      }

      if (string(repl)->len)
      {
	 memcpy(p, string(repl)->str, string(repl)->len) ;
	 p += string(repl)->len ;
      }
      if (back->len)  memcpy(p, back->str, back->len) ;
   }

   /* cleanup, repl is freed by the caller */
   mawk_repl_destroy(MAWK, &xrepl) ;
   free_STRING(back) ;

   return ret_val ;
}

/* set up for call to gsub() */
CELL *
mawk_bi_gsub(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
   CELL *cp ;			 /* pts at the replacement target */
   CELL sc ;			 /* copy of replacement target */
   CELL tc ;			 /* build the result here */

   sp -= 2 ;
   if (sp->type != C_RE)  mawk_cast_to_RE(MAWK, sp) ;
   if ((sp + 1)->type != C_REPL && (sp + 1)->type != C_REPLV)
      mawk_cast_to_REPL(MAWK, sp + 1) ;

   mawk_cellcpy(MAWK, &sc, cp = (CELL *) (sp + 2)->ptr) ;
   if (sc.type < C_STRING)  mawk_cast1_to_s(MAWK, &sc) ;

   MAWK->repl_cnt = 0 ;
   tc.ptr = (PTR) gsub(MAWK,sp->ptr, sp + 1, string(&sc)->str, 1) ;

   if (MAWK->repl_cnt)
   {
      tc.type = C_STRING ;
      mawk_slow_cell_assign(MAWK, cp, &tc) ;
   }

   /* cleanup */
   free_STRING(string(&sc)) ; free_STRING(string(&tc)) ;
   mawk_repl_destroy(MAWK, sp + 1) ;

   sp->type = C_DOUBLE ;
   sp->dval = (double) MAWK->repl_cnt ;
   return sp ;
}


CELL *
mawk_bi_call(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
  int retval ;
	char *s;
	int i, numargs;
	double d;
	SYMTAB *fs;
	const char *fn;
	CELL *ret;

	numargs = sp->type;
	sp -= numargs;
  if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp);
  fn = string(sp)->str;

	fs = mawk_find(MAWK, fn, 0);

	if ((fs == NULL) || (fs->type != ST_FUNCT)) {
		/* does not exist or not a function */
		if (fs == NULL)
			mawk_set_errno(MAWK, "1 object does not exist");
		else
			mawk_set_errno(MAWK, "2 object is not a function");

		for(i = 0; i < numargs; i++) {
			cell_destroy(&sp[i]);
			sp[i].type = C_NOINIT;
			sp[i].dval = 0;
		}
		sp[numargs].type = C_NOINIT;
		sp[numargs].dval = 0;

		return sp;
	}

	mawk_set_errno(MAWK, "");
	ret = mawk_call(MAWK, sp+numargs-1, fs->stval.fbp, numargs-1);
	cell_destroy(sp);
	*sp = *ret;
	return sp;
}

CELL *
mawk_bi_acall(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
  int retval ;
	char *s;
	int i, numargs, n;
	double d;
	SYMTAB *fs, *fa;
	const char *fn;
	CELL *arr, *ret, *elem, idx;

	numargs = 1;
	sp -= numargs;

  if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp);
  fn = string(sp)->str;

	fs = mawk_find(MAWK, fn, 0);

	if (fs == NULL) {
		/* does not exist */
		mawk_set_errno(MAWK, "1 object does not exist");
		sp++;
		goto error;
	}

	sp++;
  if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp);

	fa = mawk_find(MAWK, string(sp)->str, 0);
	if (fa == NULL) {
		mawk_set_errno(MAWK, "3 argument array does not exist");
		goto error;
	}

	if (fa->type != ST_ARRAY) {
		mawk_set_errno(MAWK, "4 argument array is not an array");
		goto error;
	}

	for(numargs = 1;; numargs++) {
    inc_sp() ;
    idx.type= C_DOUBLE;
    idx.dval = numargs;
		elem = mawk_array_find(MAWK, fa->stval.array, &idx, 0);
		if (elem == NULL)
			break;
    mawk_cellcpy(MAWK, sp, elem) ;
	}

	if (fs->type == ST_FUNCT) {
		ret = mawk_call(MAWK, sp, fs->stval.fbp, numargs);
		mawk_set_errno(MAWK, "");
	}
	else {
		mawk_set_errno(MAWK, "5 object is a function");
		idx.type = C_NOINIT;
		idx.dval = 0;
		ret = &idx;
	}

	sp-=numargs;
	sp--;

	cell_destroy(sp);
	*sp = *ret;
	return sp;

error:;
	cell_destroy(sp);
	sp--;
	sp->type = C_NOINIT;
	return sp;
}

CELL *
mawk_bi_valueof(MAWK, sp)
   mawk_state_t *MAWK;
   register CELL *sp ;
{
  int retval ;
	char *s;
	int i, numargs, n;
	double d;
	SYMTAB *fs, *fa;
	CELL *arr, *ret, *idx;

	numargs = sp->type;
	sp -= numargs;

  if (sp->type < C_STRING)  mawk_cast1_to_s(MAWK, sp);

	fs = mawk_find(MAWK, string(sp)->str, 0);

	if (fs == NULL) {
		/* does not exist */
		mawk_set_errno(MAWK, "1 object does not exist");
		goto error;
	}

	if (fs->type == ST_FUNCT) {
		mawk_set_errno(MAWK, "5 object is a function");
		goto error;
	}

	if (numargs > 1) {
		if (fs->type == ST_ARRAY) {
			idx = sp+1;
			ret = mawk_array_find(MAWK, fs->stval.array, idx, 0);
		}
		else {
			mawk_set_errno(MAWK, "6 object is not an array, can not index it");
			goto error;
		}
	}
	else {
		if (fs->type == ST_ARRAY) {
			mawk_set_errno(MAWK, "7 object is an array, need to index it");
			goto error;
		}
		ret = fs->stval.cp;
	}

	if ((ret != NULL) && ((ret->type == ST_VAR) || (ret->type == ST_KEYWORD))) {
		mawk_set_errno(MAWK, "");

		cell_destroy(sp);
		*sp = *ret;
		return sp;
	}

	if (ret != NULL)
		mawk_set_errno(MAWK, "8 object is not a variable (and not a function either)");

error:;
	cell_destroy(sp);
	sp->type = C_NOINIT;
	return sp;
}
