
/********************************************
code.c
copyright 1991-93, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/


/* $Log: code.c,v $
 * Revision 1.6  1995/06/18  19:42:13  mike
 * Remove some redundant declarations and add some prototypes
 *
 * Revision 1.5  1995/06/09  23:21:36  mike
 * make sure there is an execution block in case user defines function,
 * but no pattern-action pairs
 *
 * Revision 1.4  1995/03/08  00:06:22  mike
 * add a pointer cast
 *
 * Revision 1.3  1994/10/08  19:15:29  mike
 * remove SM_DOS
 *
 * Revision 1.2  1993/07/07  00:07:38  mike
 * more work on 1.2
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:10  mike
 * move source to cvs
 *
 * Revision 5.4	 1993/01/14  13:11:11  mike
 * code2() -> xcode2()
 *
 * Revision 5.3	 1993/01/09  20:15:35  mike
 * code_pop checks if the resolve_list needs relocation
 *
 * Revision 5.2	 1993/01/07  02:50:33  mike
 * relative vs absolute code
 *
 * Revision 5.1	 1991/12/05  07:55:43  brennan
 * 1.1 pre-release
 *
*/

/*  code.c  */

#include "mawk.h"
#include "code.h"
#include "init.h"
#include "jmp.h"
#include "field.h"


static CODEBLOCK *PROTO(new_code, (mawk_state_t *)) ;

/* grow the active code */
void
mawk_code_grow(mawk_state_t *MAWK)
{
   unsigned oldsize = mawk_code_limit - mawk_code_base ;
   unsigned newsize = PAGESZ + oldsize ;
   unsigned delta = mawk_code_ptr - mawk_code_base ;

   if (mawk_code_ptr > mawk_code_limit)  mawk_bozo(MAWK, "CODEWARN is too small") ;

   mawk_code_base = (INST *)
      mawk_zrealloc(MAWK, mawk_code_base, INST_BYTES(oldsize),
	       INST_BYTES(newsize)) ;
   mawk_code_limit = mawk_code_base + newsize ;
   mawk_code_warn = mawk_code_limit - CODEWARN ;
   mawk_code_ptr = mawk_code_base + delta ;
}

/* shrinks executable code that's done to its final size */
INST *
mawk_code_shrink(MAWK, p, sizep)
   mawk_state_t *MAWK;
   CODEBLOCK *p ;
   unsigned *sizep ;
{

   unsigned oldsize = INST_BYTES(p->limit - p->base) ;
   unsigned newsize = INST_BYTES(p->ptr - p->base) ;
   INST *retval ;

   *sizep = newsize ;

   retval = (INST *) mawk_zrealloc(MAWK, p->base, oldsize, newsize) ;
   ZFREE(p) ;
   return retval ;
}


/* code an op and a pointer in the active_code */
void
mawk_xcode2(MAWK, op, ptr)
   mawk_state_t *MAWK;
   int op ;
   PTR ptr ;
{
   register INST *p = mawk_code_ptr + 2 ;

   if (p >= mawk_code_warn)
   {
      mawk_code_grow(MAWK) ;
      p = mawk_code_ptr + 2 ;
   }

   p[-2].op = op ;
   p[-1].ptr = ptr ;
   mawk_code_ptr = p ;
}

/* code two ops in the active_code */
void
mawk_code2op(MAWK, x, y)
    mawk_state_t *MAWK;
   int x, y ;
{
   register INST *p = mawk_code_ptr + 2 ;

   if (p >= mawk_code_warn)
   {
      mawk_code_grow(MAWK) ;
      p = mawk_code_ptr + 2 ;
   }

   p[-2].op = x ;
   p[-1].op = y ;
   mawk_code_ptr = p ;
}

void
mawk_code_init(MAWK)
mawk_state_t *MAWK;
{
   MAWK->main_code_p = new_code(MAWK) ;

   MAWK->active_code = *MAWK->main_code_p ;
   code1(_OMAIN) ;
}

/* final code relocation
   set_code() as in set concrete */
void
mawk_set_code(mawk_state_t *MAWK)
{
   /* set the main code which is active_code */
   if (MAWK->end_code_p || mawk_code_offset > 1)
   {
      int gl_offset = mawk_code_offset ;

      if (MAWK->NR_flag)  mawk_code2op(MAWK, OL_GL_NR, _HALT) ;
      else  mawk_code2op(MAWK, OL_GL, _HALT) ;

      *MAWK->main_code_p = MAWK->active_code ;
      MAWK->main_start = mawk_code_shrink(MAWK, MAWK->main_code_p, &MAWK->main_size) ;
      MAWK->next_label = MAWK->main_start + gl_offset ;
      MAWK->execution_start = MAWK->main_start ;
   }
   else	 /* only BEGIN */
   {
      zfree(MAWK, mawk_code_base, INST_BYTES(PAGESZ)) ;
      ZFREE(MAWK->main_code_p) ;
   }

   /* set the END code */
   if (MAWK->end_code_p)
   {
      unsigned dummy ;

      MAWK->active_code = *MAWK->end_code_p ;
      mawk_code2op(MAWK, _EXIT0, _HALT) ;
      *MAWK->end_code_p = MAWK->active_code ;
      MAWK->end_start = mawk_code_shrink(MAWK, MAWK->end_code_p, &dummy) ;
   }

   /* set the BEGIN code */
   if (MAWK->begin_code_p)
   {
      MAWK->active_code = *MAWK->begin_code_p ;
      if (MAWK->main_start)  mawk_code2op(MAWK, _JMAIN, _HALT) ;
      else  mawk_code2op(MAWK, _EXIT0, _HALT) ;
      *MAWK->begin_code_p = MAWK->active_code ;
      MAWK->begin_start = mawk_code_shrink(MAWK, MAWK->begin_code_p, &MAWK->begin_size) ;

      MAWK->execution_start = MAWK->begin_start ;
   }

   if ( ! MAWK->execution_start )
   {
      /* program had functions but no pattern-action bodies */
      MAWK->execution_start = MAWK->begin_start = (INST*) zmalloc(MAWK, 2*sizeof(INST)) ;
      MAWK->execution_start[0].op = _EXIT0 ;
      MAWK->execution_start[1].op = _HALT  ;
   }
}

void
mawk_dump_code(mawk_state_t *MAWK)
{
   mawk_fdump(MAWK) ;			 /* dumps all user functions */
   if (MAWK->begin_start)  
   { fprintf(stdout, "BEGIN\n") ; 
     mawk_da(MAWK, MAWK->begin_start, stdout) ; }
   if (MAWK->end_start)  
   { fprintf(stdout, "END\n") ; 
     mawk_da(MAWK, MAWK->end_start, stdout) ; }
   if (MAWK->main_start)  
   { fprintf(stdout, "MAIN\n") ; 
     mawk_da(MAWK, MAWK->main_start, stdout) ; }
}


static CODEBLOCK *
new_code(MAWK)
   mawk_state_t *MAWK;
{
   CODEBLOCK *p = ZMALLOC(CODEBLOCK) ;

   p->base = (INST *) zmalloc(MAWK, INST_BYTES(PAGESZ)) ;
   p->limit = p->base + PAGESZ ;
   p->warn = p->limit - CODEWARN ;
   p->ptr = p->base ;

   return p ;
}

/* moves the active_code from MAIN to a BEGIN or END */

void
mawk_be_setup(MAWK, scope)
   mawk_state_t *MAWK;
   int scope ;
{
   *MAWK->main_code_p = MAWK->active_code ;

   if (scope == SCOPE_BEGIN)
   {
      if (!MAWK->begin_code_p)  MAWK->begin_code_p = new_code(MAWK) ;
      MAWK->active_code = *MAWK->begin_code_p ;
   }
   else
   {
      if (!MAWK->end_code_p)	MAWK->end_code_p = new_code(MAWK) ;
      MAWK->active_code = *MAWK->end_code_p ;
   }
}
