
/********************************************
re_cmpl.c
copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/* $Log: re_cmpl.c,v $
 * Revision 1.6  1994/12/13  00:14:58  mike
 * \\ -> \ on second replacement scan
 *
 * Revision 1.5  1994/10/08  19:15:51  mike
 * remove SM_DOS
 *
 * Revision 1.4  1993/07/21  01:17:53  mike
 * handle "&" as replacement correctly
 *
 * Revision 1.3  1993/07/17  13:23:10  mike
 * indent and general code cleanup
 *
 * Revision 1.2  1993/07/15  23:38:23  mike
 * SIZE_T and indent
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:19  mike
 * move source to cvs
 *
 * Revision 5.2	 1993/01/01  21:30:48  mike
 * split mawk_new_STRING() into mawk_new_STRING and mawk_new_STRING0
 *
 * Revision 5.1	 1991/12/05  07:56:25  brennan
 * 1.1 pre-release
 *
*/


/*  re_cmpl.c  */

#include "mawk.h"
#include "memory.h"
#include "scan.h"
#include "regexp.h"
#include "repl.h"


static CELL *PROTO(REPL_compile, (mawk_state_t *,STRING *)) ;

static char efmt[] = "regular expression compile failed (%s)\n%s" ;

/* compile a STRING to a regular expression machine.
   Search a list of pre-compiled strings first
*/
PTR
mawk_re_compile(MAWK, sval)
   mawk_state_t *MAWK;
   STRING *sval ;
{
   register RE_NODE *p ;
   RE_NODE *q ;
   char *s ;

   /* search list */
   s = sval->str ;
   p = MAWK->re_list ;
   q = (RE_NODE *) 0 ;
   while (p)
   {
      if (strcmp(s, p->sval->str) == 0) /* found */
      {
	 if (!q)		/* already at front */
	    goto _return ;
	 else  /* delete from list for move to front */
	 {
	    q->link = p->link ; goto found ; 
	 }

      }
      else
      {
	 q = p ; p = p->link ; 
      }
   }

   /* not found */
   p = ZMALLOC(RE_NODE) ;
   p->sval = sval ;

   sval->ref_cnt++ ;
   if (!(p->re = mawk_REcompile(s)))
   {
      if (MAWK->mawk_state == EXECUTION)
	 mawk_rt_error(MAWK, efmt, REerrlist[REerrno], s) ;
      else  /* compiling */
      {
	 mawk_compile_error(MAWK, efmt, REerrlist[REerrno], s) ;
	 return (PTR) 0 ;
      }
   }


found :
/* mawk_insert p at the front of the list */
   p->link = MAWK->re_list ; MAWK->re_list = p ;

_return :

#ifdef	DEBUG
   if (MAWK->dump_RE)	 REmprint(p->re, stderr) ;
#endif
   return p->re ;
}



/* this is only used by mawk_da() */


char *
mawk_re_uncompile(MAWK, m)
   mawk_state_t *MAWK;
   PTR m ;
{
   register RE_NODE *p ;

   for (p = MAWK->re_list; p; p = p->link)
      if (p->re == m)  return p->sval->str ;
#ifdef	DEBUG
   mawk_bozo(MAWK, "non compiled machine") ;
#endif
	return NULL;
}



/*=================================================*/
/*  replacement	 operations   */

/* create a replacement CELL from a STRING *  */

static CELL *
REPL_compile(MAWK, sval)
   mawk_state_t *MAWK;
   STRING *sval ;
{
   int i = 0 ;
   register char *p = sval->str ;
   register char *q ;
   char *xbuff ;
   CELL *cp ;

   q = xbuff = (char *) zmalloc(MAWK, sval->len + 1) ;

   while (1)
   {
      switch (*p)
      {
	 case 0:
	    *q = 0 ;
	    goto done ;

	 case '\\':
	    if (p[1] == '&'|| p[1] == '\\')
	    {
	       *q++ = p[1] ;
	       p += 2 ;
	       continue ;
	    }
	    else  break ;

	 case '&':
	    /* if empty we don't need to make a node */
	    if (q != xbuff)
	    {
	       *q = 0 ;
	       split_buff[i++] = mawk_new_STRING(MAWK, xbuff) ;
	    }
	    /* and a null node for the '&'  */
	    split_buff[i++] = (STRING *) 0 ;
	    /*	reset  */
	    p++ ;  q = xbuff ;
	    continue ;

	 default:
	    break ;
      }

      *q++ = *p++ ;
   }

done :
   /* if we have one empty string it will get made now */
   if (q > xbuff || i == 0)  split_buff[i++] = mawk_new_STRING(MAWK, xbuff) ;

   /* This will never happen */
   if (i > MAX_SPLIT)  mawk_overflow(MAWK, "replacement pieces", MAX_SPLIT) ;

   cp = ZMALLOC(CELL) ;
   if (i == 1 && split_buff[0])
   {
      cp->type = C_REPL ;
      cp->ptr = (PTR) split_buff[0] ;
   }
   else
   {
      STRING **sp = (STRING **)
      (cp->ptr = zmalloc(MAWK, sizeof(STRING *) * i)) ;
      int j = 0 ;

      while (j < i)  *sp++ = split_buff[j++] ;

      cp->type = C_REPLV ;
      cp->vcnt = i ;
   }
   zfree(MAWK, xbuff, sval->len + 1) ;
   return cp ;
}

/* free memory used by a replacement CELL  */

void
mawk_repl_destroy(MAWK, cp)
   mawk_state_t *MAWK;
   register CELL *cp ;
{
   register STRING **p ;
   unsigned cnt ;

   if (cp->type == C_REPL)  free_STRING(string(cp)) ;
   else	 /* an C_REPLV	      */
   {
      p = (STRING **) cp->ptr ;
      for (cnt = cp->vcnt; cnt; cnt--)
      {
	 if (*p)  free_STRING(*p) ;
	 p++ ;
      }
      zfree(MAWK, cp->ptr, cp->vcnt * sizeof(STRING *)) ;
   }
}

/* copy a C_REPLV cell to another CELL */

CELL *
mawk_replv_cpy(MAWK, target, source)
   mawk_state_t *MAWK;
   CELL *target, *source ;
{
   STRING **t, **s ;
   unsigned cnt ;

   target->type = C_REPLV ;
   cnt = target->vcnt = source->vcnt ;
   target->ptr = (PTR) zmalloc(MAWK, cnt * sizeof(STRING *)) ;

   t = (STRING **) target->ptr ;
   s = (STRING **) source->ptr ;
   while (cnt)
   {
      cnt-- ;
      if ( *s ) (*s)->ref_cnt++ ;
      *t++ = *s++ ;
   }
   return target ;
}


/* search the list (with move to the front) for a compiled
   separator.
   return a ptr to a CELL (C_REPL or C_REPLV)
*/

CELL *
mawk_repl_compile(MAWK, sval)
   mawk_state_t *MAWK;
   STRING *sval ;
{
   register REPL_NODE *p ;
   REPL_NODE *q ;
   char *s ;

   /* search the list */
   s = sval->str ;
   p = MAWK->repl_list ;
   q = (REPL_NODE *) 0 ;
   while (p)
   {
      if (strcmp(s, p->sval->str) == 0) /* found */
      {
	 if (!q)		/* already at front */
	    return p->cp ;
	 else  /* delete from list for move to front */
	 {
	    q->link = p->link ;
	    goto found ;
	 }

      }
      else
      {
	 q = p ; p = p->link ; 
      }
   }

   /* not found */
   p = ZMALLOC(REPL_NODE) ;
   p->sval = sval ;
   sval->ref_cnt++ ;
   p->cp = REPL_compile(MAWK, sval) ;

found :
/* mawk_insert p at the front of the list */
   p->link = MAWK->repl_list ; MAWK->repl_list = p ;
   return p->cp ;
}

/* return the string for a CELL or type REPL or REPLV,
   this is only used by mawk_da()  */


char *
mawk_repl_uncompile(MAWK, cp)
   mawk_state_t *MAWK;
   CELL *cp ;
{
   register REPL_NODE *p = MAWK->repl_list ;

   if (cp->type == C_REPL)
   {
      while (p)
      {
	 if (p->cp->type == C_REPL && p->cp->ptr == cp->ptr)
	    return p->sval->str ;
	 else  p = p->link ;
      }
   }
   else
   {
      while (p)
      {
	 if (p->cp->type == C_REPLV &&
	     memcmp(cp->ptr, p->cp->ptr, cp->vcnt * sizeof(STRING *))
	     == 0)
	    return p->sval->str ;
	 else  p = p->link ;
      }
   }

#if  DEBUG
   mawk_bozo(MAWK, "unable to uncompile an repl") ;
#endif
	return  NULL;
}

/*
  convert a C_REPLV to	C_REPL
     replacing the &s with sval
*/

CELL *
mawk_replv_to_repl(MAWK, cp, sval)
mawk_state_t *MAWK;
CELL *cp ; STRING *sval ;
{
   register STRING **p ;
   STRING **sblock = (STRING **) cp->ptr ;
   unsigned cnt, vcnt = cp->vcnt ;
   unsigned len ;
   char *target ;

#ifdef	DEBUG
   if (cp->type != C_REPLV)  mawk_bozo(MAWK, "not replv") ;
#endif

   p = sblock ; cnt = vcnt ; len = 0 ;
   while (cnt--)
   {
      if (*p)  len += (*p++)->len ;
      else
      {
	 *p++ = sval ;
	 sval->ref_cnt++ ;
	 len += sval->len ;
      }
   }
   cp->type = C_REPL ;
   cp->ptr = (PTR) mawk_new_STRING0(MAWK, len) ;

   p = sblock ; cnt = vcnt ; target = string(cp)->str ;
   while (cnt--)
   {
      memcpy(target, (*p)->str, (*p)->len) ;
      target += (*p)->len ;
      free_STRING(*p) ;
      p++ ;
   }

   zfree(MAWK, sblock, vcnt * sizeof(STRING *)) ;
   return cp ;
}
